package EPIC;

###################################################################################################
#
# Moduel name: EPIC.pm
#
# This module is used to sort and select for the EPIC markers.
#
# Datafile: onehit file has the list of single-copy CDSs that have one hit in all ref sequences.
#
# Input: query taxon name.
#
# Output: a list of EPIC markers and files of aligned sequences.
#
# Written by 
#                 Chenhong Li
#                 Univeristy of Nebraska - Lincoln, USA.
#                 Created on Dec. 2008
#                 Last modified at
#
###################################################################################################

use strict;
use warnings;



###################################################################################################
#
# sub routine: selectEPIC
#
# This sub is used to screen for EPIC (size < $Li in at least one taxon).
#
# Datafile: onehit file has the list of single-copy CDSs that have one hit in all ref sequences.
#
# Imput: the name of query, reference and $Li.
#
# Output: list of EPIC markers meet the criteria.
#
###################################################################################################

sub selectEPIC {

    my $queryname = shift (@_);#get the query name
    
    my $reference = shift (@_);#get the reference name, make the array of all taxa
    my @taxa = split (/\s/, $reference);
    my @sortedtaxa = sort @taxa;
    unshift @sortedtaxa, $queryname;#the taxa list always starts with query, then sorted ref
    
    my $Li = shift (@_); #get the minimum intron size
    
    my $infile = "../$queryname.results/$queryname.onehitCDSmarkers.txt";
    my $outfile = "../$queryname.results/$queryname.EPIClist.txt";
    
    my ($IN_FILE, $OUT_FILE);

    open ($IN_FILE, "<$infile") or die "Can't open the onehitCDSmarkers.txt file!";
    open ($OUT_FILE, ">$outfile") or die "Can't open the output file to save EPIC";
    
    #skip the first line and print the title line to new file
    readline($IN_FILE);
    print $OUT_FILE "EPIC_ID\t";
    foreach my $taxon (@sortedtaxa) {
        print $OUT_FILE "$taxon\t";
    }
    print $OUT_FILE "Average Identity\tLocus_tag|Strand|Note\n";


    my $numberofEPIC = 0;
    my $EPIClabel = 0;
    my ($geneid, $geneidlag, $genesum, $identity, $identitylag);
    $geneidlag = "id";#assign an initial value to $geneidlag
    my (%position, %positionlag);

    
    while (my $line = readline($IN_FILE)) {

        chomp $line;
        my @column = split (/\t/, $line);#split columns

        my $id = shift @column;#get the gene informtion
        ($geneid) = $id =~ /[\w:]+\|(\w+)\|.+/;
        ($genesum) = $id =~ /[\w:]+\|\w+\|(.+)/;
        my ($minusstrain) = $genesum =~ /.*\|(.+)\|.*/;
        my $newline = "\t";
        foreach my $taxon (@sortedtaxa) {
            $position{$taxon} = shift @column;#get the global position for each taxa for this gene
        }
        $identity = shift @column;#get identity
        
        if ($geneid eq $geneidlag) {
            my $meanidentity = ($identity + $identitylag)/2;
            my $isEPIC = 0;
            foreach my $taxon (@sortedtaxa) {
                my $intronsize = &EPIC::intronsize("$position{$taxon}", "$positionlag{$taxon}");#calculate the intron size
                my $newentry;
                
                if ($minusstrain eq -1) { #if it is minus strain
                    $newentry = $position{$taxon} . "\|" . $positionlag{$taxon} . "\|" . $intronsize . "\t";
                }
                elsif ($minusstrain eq 1) { #if it is plus strain
                    $newentry = $positionlag{$taxon} . "\|" . $position{$taxon} . "\|" . $intronsize . "\t";
                }
                $newline .= $newentry;
                $isEPIC = 1 if ($intronsize <= $Li);#if any taxon has intron size < $Li
                if ($intronsize == 0) {#if the gene structure doesn't hold among taxa
                    $isEPIC = 0;
                    last;
                }
            }

            if ($isEPIC) {#save EPIC to file
                $numberofEPIC ++;
                $EPIClabel ++;#affix to distinguish different EPIC of the same gene
                my $epicid = "$geneid" . "_EPIC" . $EPIClabel;
                print $OUT_FILE "$epicid";
                print $OUT_FILE "$newline";
                print $OUT_FILE "$meanidentity\t" . $genesum . "\n";
            }
            
        }
        else {
            $EPIClabel = 0;#reset number of EPIC in each gene to zero
        }
       
        #assign the current entry to the lag
        foreach my $taxon (@sortedtaxa) {
            $positionlag{$taxon} = $position{$taxon};
        }
        $geneidlag = $geneid;
        $identitylag = $identity;
        

    }

    
    close ($IN_FILE) or die "Can't close  the onehitCDSmarkters.txt file!";
    close ($OUT_FILE) or die "Can't close the output file";





    return($numberofEPIC);

}




###################################################################################################
#
# sub routine: intronsize
#
# This sub is used to calculate intron size between two CDS.
#
# Imput: the global (chromosomal) position of two flanking CDS.
#
# Output: list of EPIC markers meet the criteria.
#
###################################################################################################

sub intronsize {

    my $cdsone = shift (@_);#get the postition for the first CDS
    my $cdstwo = shift (@_);#get the position for the second CDS
    
    my $intronsize;
    
    my (@cds1, @cds2, @pos); 
    @cds1 = split (/:/, $cdsone);#get the start and stop position for each CDS
    @cds2 = split (/:/, $cdstwo);
    
    if ($cds1[0] eq $cds2[0]) {#calculate the intron size between two CDS
        push @pos, $cds1[1], $cds1[2], $cds2[1], $cds2[2];
        my @sortedpos = sort {$a <=> $b} @pos;
        $intronsize = $sortedpos[2] - $sortedpos[1] - 1;
    }
    else {
        $intronsize = 0;#if two CDS are not from the same chromosome, use 0 to indicate the problem
    }

    return($intronsize);

}




###################################################################################################
#
# sub routine: EPICseqfile
#
# This sub is used to print the fasta files for EPIC markers, two concatenated flanking CDS
# seperated by a number indicating the size of the intron.
#
# Datafile: list of EPIC.
#
# Imput: the name of query, $pid.
#
# Output: fasta files of EPIC.
#
###################################################################################################

sub EPICseqfile {

    my $queryname = shift (@_);#get the query name
    my $pid = shift (@_);#if the average identity > $pid, then print fasta file

    my $infile = "../$queryname.results/$queryname.EPIClist.txt";
    my $IN_FILE;
    open ($IN_FILE, "<$infile") or die "Can't open the EPIClist.txt file!";
    
    
    my $titleline;
    $titleline = readline($IN_FILE);
    chomp $titleline;
    my @taxa = split(/\t/, $titleline);#get the array of taxa;
    shift @taxa;#get rid of the first and two last columns
    pop @taxa;
    pop @taxa;
    
    my %EPIC;
    
    #creat hash storing information about the EPIC markers
    while (my $line = readline($IN_FILE)) {
        chomp $line;
        
        my @entry = split(/\t/, $line);#split the entry into array
        my $taxacount = 0;
        my $numberoftaxa = scalar(@taxa);
        foreach my $taxon (@taxa) {#assign the EPIC information into the EPIC hash
            $taxacount ++;
            $EPIC{$entry[0]}{$taxon} = $entry[$taxacount];#assign the genome position into hash
            $EPIC{$entry[0]}{Identity} = $entry[$numberoftaxa + 1];#assign the average identity into hash
       }
        
    }

    close ($IN_FILE) or die "Can't close  the EPIClist.txt file!";

    
    
    
    #make fasta files for selected EPIC markers
    my @EPICid = keys %EPIC;#get the id from the hash
    my @sortedid = sort @EPICid;
    
    my $dir = "../$queryname.results/" . $queryname . "_EPICmarkers";#make a director for storing the fasta files
    `mkdir $dir -p`;
    
    my $epiccount = 0;
    foreach my $id (@sortedid){#go through all EPIC
        if ($EPIC{$id}{Identity} > $pid) {
        
            $epiccount ++;
            my $outfile = "$dir\/$id.fas";
            my $OUT_FILE;
            open ($OUT_FILE, ">$outfile") or die "Can't open the output file to save EPIC";
       
            foreach my $taxon (@taxa) {#read information from the EPIC hash
                my $temp = $EPIC{$id}{$taxon};
                my @epicinfo = split(/\|/, $temp);
                my $intronsize = $epicinfo[2];
                my @pos = split(/\:/, $epicinfo[0]);
                my @posnext = split(/\:/, $epicinfo[1]);

                my $seqfile = "../$taxon.db/$taxon.genome.new.fas";
                my $indexfile = "../$taxon.db/$taxon.genome.new.index";

                my $seqid = $pos[0];
                my $pos1 = $pos[1];
                my $pos2 = $pos[2];
                my ($subseq, $strain) = &makequery::extractseq($seqid, $pos1, $pos2, $seqfile, $indexfile);
                print $OUT_FILE ">$taxon\n";
                print $OUT_FILE "$subseq";
            
                print $OUT_FILE "XXXXX";#intron indicator for PriFinder primer designing
 
                $seqid = $posnext[0];
                $pos1 = $posnext[1];
                $pos2 = $posnext[2];
                ($subseq, $strain) = &makequery::extractseq($seqid, $pos1, $pos2, $seqfile, $indexfile);
                print $OUT_FILE "$subseq";
            
                print $OUT_FILE "$intronsize\n";
            }
           
            close ($OUT_FILE) or die "Can't close the output file";
        
        }
        
        else {
            next;
        }
       
        
    
    
    }



    return($epiccount);

}





1