package blastandparse;

###################################################################################################
#
# Moduel name: blastandparse.pm
#
# This module is used to do blast and parse the results, two subs: withingenome and betweengenome
# used to parse the result of blast within genomes and between genomes respectively
#
# Written by 
#                 Chenhong Li
#                 Univeristy of Nebraska - Lincoln, USA.
#                 Created on Oct. 2008
#                 Last modified at
#
###################################################################################################

use strict;
use warnings;



###################################################################################################
#
# sub routine: within
#
# This sub is used to screen single-copy CDS within genome.
#
# Datafile: query fasta sequences and the local blast data base.
#
# Imput: the name of query, $Iwg, $Cwg, $E.
#
# Output: list of "single-copy" CDS meet the criteria.
#
###################################################################################################

sub within {

    my $query = shift(@_); #get the query name from the main program
    my $Iwg = shift(@_); #get identity parameter from the main program
    my $Cwg = shift(@_); #get the coverage parameter from the main program
    my $E = shift(@_); #get the E-value for blast from the main program

    #call outside program blastn
    `/arch/bin/blastall -p blastn -i "../$query.results/$query.large_cds.fas" -d "../$query.db/$query.db" -o "../$query.results/$query.withinblast.txt" -q -1 -e $E -m 8`;



    #parse blast results, screen for single-copy CDS
    my ($IN_FILE, $OUT_FILE);
    my ($infilename, $outfilename);

    my @column;
    my $numberofhits = 0;
    my $cdsname = "";
    my $cdsid;
    my $numberofsingles = 0;
    my ($line, $linelag);

    $infilename = "../$query.results/$query.withinblast.txt"; #get the name of blast results file
    $outfilename = "../$query.results/$query.single_cds.txt"; #make the name of outfile
    open ($IN_FILE, "<$infilename") or die "Can't open the input file!!!";
    open ($OUT_FILE, ">$outfilename") or die "Can't open the output file!!!";#output to outfile


    while ($line = readline($IN_FILE)) {

        chomp $line;
        @column = split (/\t/, $line);#split columns

        my ($cdsstart, $cdsend) = $column[0] =~ /.+:(\d+):(\d+)\|.+/;
        my $cdslength = abs($cdsend - $cdsstart) + 1; #calculate the lenght of CDS

        my $coverage = 100*($column[3]/$cdslength); #calculate the coverage
        my $identity = $column[2]; #get the identity
        
        if (($identity>= $Iwg) && ($coverage >= $Cwg)) { #if the entries meeting requrement
            if ($column[0] eq $cdsname) { #if the same cds
                $numberofhits ++;
            } else {
                if ($numberofhits == 1) { #if only one hit
                    print $OUT_FILE "$linelag\n";
                    $numberofsingles ++; #number of single-copy CDS
                }
                $numberofhits = 1; #reset the $numberofhits
                $cdsname = $column[0]; #assign the next CDS name
                $linelag = $line;
            }
        }
    }

    #print the last entry, if it meet the criteria
    if ($numberofhits == 1) { #if only one hit
        print $OUT_FILE "$linelag\n";
        $numberofsingles ++; #number of single-copy cds
    }


    close ($IN_FILE) or die "Can't close the input file!!!";
    close ($OUT_FILE) or die "Can't close the output file!!!";

    return($numberofsingles);
}



###################################################################################################
#
# sub routine: singleseq
#
# This sub is used to get the CDS fasta squences.
#
# Imput data include the name of taxon.
#
# Output is fasta file of selected CDS.
#
###################################################################################################

sub singleseq {

    my $query = shift(@_); #get the query name from the main program


    my ($CDS_FILE, $FASTA_FILE, $OUT_FILE);
    my ($cdsfilename, $fastafilename, $outfilename);

    $cdsfilename = "../$query.results/$query.single_cds.txt"; #get the name of CDS file
    $fastafilename = "../$query.results/$query.large_cds.fas"; #get the name of seq file
    $outfilename = "../$query.results/$query.single_cds.fas"; #make the name of outfile
    open ($CDS_FILE, "<$cdsfilename") or die "Can't open the single CDS file!!!";
    open ($OUT_FILE, ">$outfilename") or die "Can't open the output file!!!";#output to outfile


    while (my $line = readline($CDS_FILE)) { #read the name of single-copy CDS
        my @column = split (/\|/, $line);#split columns
        my $seqid = ">" . $column[0];

        open ($FASTA_FILE, "<$fastafilename") or die "Can't open the fasta file!!!";

        while (my $lines = readline($FASTA_FILE)) {
            chomp $lines;
            if ($lines =~ /^$seqid/) { #if match the seq id
                print $OUT_FILE "$lines\n";
                my $seq = readline($FASTA_FILE); #the next line is the sequence
                print $OUT_FILE "$seq";
                last;
            } 
            else {
                next;
            }
        }

        close ($FASTA_FILE) or die "Can't close the fasta file!!!";

    }


    close ($CDS_FILE) or die "Can't close the single CDS file!!!";
    close ($OUT_FILE) or die "Can't close the output file!!!";

    return(1);
}



###################################################################################################
#
# sub routine: between
#
# This sub is used to screen conserved CDS between genome.
#
# Imput data include $CDS, the name of query, the name of reference, $E, $Ibg, $Cbg.
#
# Output file is list of cds meet the criteria.
#
###################################################################################################

sub between {

    my $CDS = shift(@_); #get the control variable for CDS markers
    my $query = shift(@_); #get the query name from the main program
    my $reference = shift(@_); #get reference names from the main program
    my @refgenome = split (/\s/, $reference); #split the names of refer genomes
    my @sortedrefgenome = sort @refgenome;
    my $E = shift(@_); #get the E-value for blast from the main program
    my $Ibg = shift(@_); #get the identity parameter from the main program
    my $Ibgsec = shift(@_); #get the identity parameter from the main program
    my $Cbg = shift(@_); #get the coverage parameter from the main program
    my $Cbgsec = shift(@_); #get the coverage parameter from the main program for the second hit
    
    #call outside program blastn
    my $queryfile = "../$query.results/$query.single_cds.fas"; #get the name of blast query file
    foreach my $ref (@sortedrefgenome) {
        my $databasefile = "../$ref.db/$ref.db";
        my $blastout = "../$query.results/$query.$ref.blast.txt";
        print "Blasting, $query against $ref...\n";
        `/arch/bin/blastall -p blastn -i "$queryfile" -d "$databasefile" -o "$blastout" -q -1 -e $E -m 8`;
    }




    #screen for conserved CDS
    my %cdslist;
    my $BLAST_FILE;
    my $blastname;

    foreach my $ref (@sortedrefgenome) {
        $blastname = "../$query.results/$query.$ref.blast.txt"; #get the name of blast results file
        open ($BLAST_FILE, "<$blastname") or die "Can't open the input file!!!";

        my @column;
        my $numberofhits = 0;
        my $firstentry = "yes";
        my $line;
        my $cdslag;
        my $cdslistkey;
        my ($queryidlag, $querypos1lag, $querypos2lag, $refchromosomelag, $refpos1lag, $refpos2lag, $identitylag, $coveragelag, $cdslengthlag, $cdslagname);

        while ($line = readline($BLAST_FILE)) {
            chomp $line;
            @column = split (/\t/, $line);#split columns

            my ($cdsstart, $cdsend) = $column[0] =~ /.+:(\d+):(\d+)\|.+/;
            my $querylength = abs($cdsend - $cdsstart) + 1; #calculate the length of CDS

            my $cdslength = $column[3];#the length of query match
            my $coverage = 100*($cdslength/$querylength); #calculate the coverage
            my $identity = $column[2]; #get the identity
            
            
            # if the entry has an id% and coverage larger than the given values or they are larger than the values set for multiple hits if the $numberofhits is bigger than 1
            if ((($identity >= $Ibg) && ($coverage >= $Cbg)) || ((($numberofhits >= 1) && ($column[0] eq $cdslag)) && (($identity >= $Ibgsec) && ($coverage >= $Cbgsec)))) {
                if ($firstentry eq "yes") { #if it is the first entry
                    $numberofhits ++;
                    $firstentry = "no";
                }
                elsif ($column[0] eq $cdslag) { #if the same cds 
                    $numberofhits ++;
                } 
                else { #make the hash of selected CDS, notice the query with multiple hits only has the right information for the $numberofhits
                    $cdslist{$cdslistkey}->{$ref} = {hits => $numberofhits,
                                                    queryid => $queryidlag,
                                                    querypos1 => $querypos1lag,
                                                    querypos2 => $querypos2lag,
                                                    refchromosome => $refchromosomelag,
                                                    refpos1 => $refpos1lag,
                                                    refpos2 => $refpos2lag,
                                                    identity => $identitylag,
                                                    coverage => $coveragelag,
                                                    length => $cdslengthlag };
                    
                    $numberofhits = 1; #reset the $numberofhits
                    
                } 
   
                $cdslistkey = $column[0]; #assign the next cds name and other value of first hit
                $cdslistkey =~ s/=/ /g; #change "=" back to space
                    
                $queryidlag = $column[0];
                $querypos1lag = $column[6];
                $querypos2lag = $column[7];
                if (($refchromosomelag) = $column[1] =~ /.+:(\S+)$/) { #remove extra characters in ref id
                }
                else{
                    $refchromosomelag = $column[1];
                }
                $refpos1lag = $column[8]; 
                $refpos2lag = $column[9];
                $identitylag = $identity;
                $coveragelag = $coverage;
                $cdslengthlag = $cdslength;

                $cdslag = $column[0];

            }
            
                
        }

            #add the last entry
            $cdslist{$cdslistkey}->{$ref} = {hits => $numberofhits,
                                            queryid => $queryidlag,
                                            querypos1 => $querypos1lag,
                                            querypos2 => $querypos2lag,
                                            refchromosome => $refchromosomelag, 
                                            refpos1 => $refpos1lag, 
                                            refpos2 => $refpos2lag,
                                            identity => $identitylag,
                                            coverage => $coveragelag,
                                            length => $cdslengthlag };

        
        close ($BLAST_FILE) or die "Can't close the input file!!!";
        
    }
    
    
    
    
    
    
    #print the results of cds marker list
    my $LIST_FILE;
    my $listfilename;

    $listfilename = "../$query.results/$query.allCDSmarkers.txt"; #make the name of outfile
    open ($LIST_FILE, ">$listfilename") or die "Can't open the output file!!!";#output to outfile

    #print headline
    print $LIST_FILE "Chromosomal_position|Gene_ID|Locus_tag|Strand|Note";
    foreach my $ref (@sortedrefgenome) {
        print $LIST_FILE "\t$ref";
    }
    print $LIST_FILE "\n";
        
    
    #sort by the chromosome id then by the global position of the cds
    my @cdskey = keys (%cdslist);
    my @sortedkey = sort mysort @cdskey;
    sub mysort { 
        my $aa = $a;
        my $bb = $b;
        ($aa =~ /^\w+:\w+:(\w+):\d+:\d+\|\w+\|/)[0] cmp ($bb =~ /^\w+:\w+:(\w+):\d+:\d+\|\w+\|/)[0] || ($a =~ /^\w+:\w+:\w+:(\d+):\d+\|\w+\|/)[0] <=> ($b =~ /^\w+:\w+:\w+:(\d+):\d+\|\w+\|/)[0];
    }             


    foreach my $cdsid (@sortedkey) { #go through all cds
        
        print $LIST_FILE "$cdsid";
        
        foreach my $ref (@sortedrefgenome) { #go through all ref genome
            my $blasthits;
            if (!($cdslist{$cdsid}->{$ref}->{hits})) { #if not exist, assign zero to it
                $blasthits = 0;
                $cdslist{$cdsid}->{$ref} = {hits => 0};
            } 
            else { #if exists
                $blasthits = $cdslist{$cdsid}->{$ref}->{hits};
            }
            print $LIST_FILE "\t$blasthits";
        }
        
        print $LIST_FILE "\n";
        
    }


    close ($LIST_FILE) or die "Can't close the output file!!!";




    
    
    #select the CDS that have single hit in all ref taxa
    my @newcdskeys;
    foreach my $cdsid (@sortedkey) { #check all CDS
        my $goodcds = 1;
        foreach my $ref (@sortedrefgenome) { #check all ref, false $goodcds if hits != to 1 in any taxa
           $goodcds = 0 if ($cdslist{$cdsid}->{$ref}->{hits} != 1);
        }
        push @newcdskeys, $cdsid if ($goodcds == 1);
    }

    #print the list of cds with only one hit in all reference genomes
    my $SINGLELIST_FILE;
    my $singlelistfilename;
    $singlelistfilename = "../$query.results/$query.onehitCDSmarkers.txt"; #make the name of outfile
    open ($SINGLELIST_FILE, ">$singlelistfilename") or die "Can't open the output file!!!";#output to outfile

    #print headline
    print $SINGLELIST_FILE "Chromosomal_position|Gene_ID|Locus_tag|Strand|Note\t$query";
    foreach my $ref (@sortedrefgenome) {
        print $SINGLELIST_FILE "\t$ref";
    }
    print $SINGLELIST_FILE "\tAverage Identity\tAverage Coverage\tAverage Length\n";
        
    my $count = 0;
    foreach my $cdsid (@newcdskeys) { #go through all single cds
        $count ++;
                
        print $SINGLELIST_FILE "$cdsid";
        my ($querypos) = $cdsid =~ /\w+:\w+:([\w_\d]+:\d+:\d+)/;
        print $SINGLELIST_FILE "\t$querypos";

        my (@AI, @AC, @AL); #defind variables to calculate average identity, coverage and length
        my ($AI, $AC, $AL);
        
        foreach my $refgenome (@sortedrefgenome) { #go through all ref genome, print the position of hits

            my $refchromosome = $cdslist{$cdsid}->{$refgenome}->{refchromosome};
            my $refpos1 = $cdslist{$cdsid}->{$refgenome}->{refpos1};
            my $refpos2 = $cdslist{$cdsid}->{$refgenome}->{refpos2};
            
            
            print $SINGLELIST_FILE "\t$refchromosome:$refpos1:$refpos2";
            
            push @AI, $cdslist{$cdsid}->{$refgenome}->{identity} if ($cdslist{$cdsid}->{$refgenome}->{identity}); #if exist
            push @AC, $cdslist{$cdsid}->{$refgenome}->{coverage} if ($cdslist{$cdsid}->{$refgenome}->{coverage}); #if exist
            push @AL,  $cdslist{$cdsid}->{$refgenome}->{length}   if ($cdslist{$cdsid}->{$refgenome}->{length});   #if exist
            
        }
        
        
        $AI = &blastandparse::aver(\@AI);
        $AC = &blastandparse::aver(\@AC);
        $AL  = &blastandparse::aver(\@AL);
        

        print $SINGLELIST_FILE "\t$AI"; #print average of the array
        print $SINGLELIST_FILE "\t$AC";
        print $SINGLELIST_FILE "\t$AL";        
        print $SINGLELIST_FILE "\n";
    }


    close ($SINGLELIST_FILE) or die "Can't close the output file!!!";




    #creat fasta file for each selected CDS marker have single hit in all ref taxa
    if ($CDS ne "N") { #if search for CDS markers

        my @sortedtaxa = sort @sortedrefgenome;
        
        print "\nCreating files for selected CDS markers, please wait...\n";
    
        `mkdir ../$query.results/CDSmarkers`;

        my ($cdsname, $cdsnamelag, $idaffix, $markername);
        foreach my $newcdsid (@newcdskeys) {
            ($cdsname) = $newcdsid =~ /^[\d\w:_]+\|(\w+)\|/; # get the cds name
            
            if ($cdsnamelag && ($cdsname eq $cdsnamelag)) {
                $markername = $cdsname . $idaffix;#use a affix to the ensembl id to distinguish, if more than one markers were found in one gene
                $idaffix ++;
            }
            else {
                $markername = $cdsname;
                $idaffix = 2;#reset $idaffix
            }
            $cdsnamelag = $cdsname;
            
            my $FASTA_FILE;
            my $fastafile;
            $fastafile = "../$query.results/CDSmarkers/$markername.fas";#make file for output fasta sequences
            open ($FASTA_FILE, ">$fastafile") or die "Can't save fasta file for each CDS!";
            my $sourcefile = "../$query.results/$query.single_cds.fas";


            #extract and print out sequence for query
            my $queryseqfile = "../$query.db/$query.genome.new.fas";
            my $queryindexfile = "../$query.db/$query.genome.new.index";
            my ($queryid, $querypos1, $querypos2) = $newcdsid =~ /^\w+:\w+:(\w+):(\d+):(\d+)/;
            my ($querysubseq, $querystrain) = &makequery::extractseq($queryid, $querypos1, $querypos2, $queryseqfile, $queryindexfile);
            print $FASTA_FILE ">$query\n";
            print $FASTA_FILE "$querysubseq\n";
       
       
            foreach my $taxa (@sortedtaxa) {
                my $seqid = $cdslist{$newcdsid}->{$taxa}->{refchromosome};
                my $pos1 = $cdslist{$newcdsid}->{$taxa}->{refpos1};
                my $pos2 = $cdslist{$newcdsid}->{$taxa}->{refpos2};
                my $seqfile = "../$taxa.db/$taxa.genome.new.fas";
                my $indexfile = "../$taxa.db/$taxa.genome.new.index";
                if ($seqid) { #if defined
                    print $FASTA_FILE ">$taxa\n";
                    #extract sequence of the hit from ref genome
                    my ($subseq, $strain) = &makequery::extractseq($seqid, $pos1, $pos2, $seqfile, $indexfile);
                    print $FASTA_FILE "$subseq\n";
                }
                else {
                    next;
                }
                
            }
            close ($FASTA_FILE) or die "Can't close the fasta file!";
    
        }

    }




    
    return($count);

}




###################################################################################################
#
# sub routine: aver
#
# This sub is used to calculate the average of an array.
#
# Imput data include the array.
#
# Return is the average number.
#
###################################################################################################

sub aver {

    my $arrayref = shift(@_); #get the $seqid from main program
    my @array = @{$arrayref};
    my $sum = 0;
    my $average;
    my $numofarr = scalar(@array);
    
    foreach my $number (@array) {
        $sum = $sum + $number;
    }
    
    $average = $sum/$numofarr;
    
    return ($average);


}








1