#!/arch/bin/perl -w

###########################################################################################
#                                                                                         #
# fishintron.pl                                                                           #
#                                                                                         #
#                                                                                         #
# Data files: genome sequencies in FASTA format for both query and reference taxa (e.g.,  #
#             I downloaded ENSEMBL toplevel.fa file). The files should be saved in folder #
#             for each species separately (e.g. ../Danio_rerio.db/, you can substitute    #
#             Danio_rerio with your taxon name).                                          #
#                                                                                         #
#             genome sequencies in EMBL format for query species (e.g., I downloaded the  #
#             EMBL files from ENSEMBL). The EMBL files are used to parse information of   #
#             CDS, since the CDS information given by BioMart in ENSEMBL has errors. All  #
#             EMBL data file should be saved in the folder for each species separatly     #
#             (e.g., ../Danio_rerio.db/EMBL/, you can substitute Danio_rerio with your    #
#             taxon name).                                                                #
#                                                                                         #
#                                                                                         #
# Format database: the genome sequence file should be formated into local blast database  #
#                   using formatdb (http://www.ncbi.nlm.nih.gov/blast/docs/formatdb.html) #
#                   (e.g. formatdb -p F -i Danio_rerio.fasta -n Danio_rerio.db).          #
#                   Index files should also be made to facilitate extracting sequences    #
#                   using makeindex.pl before running the main program, see README.txt.   #
#                                                                                         #
#                                                                                         #
# Input:  the scientific name of the query species and the reference species (note, use a #
#         low dash to link the genus name and species name, separate reference taxa by a  #
#         space), Values for minimum CDS size, ($Lc), maximum intron size ($Li), blast    #
#         E-value ($E), identity within genome ($Iwg), coverage within genome ($Cwg),     #
#         identity between genome ($Ibg), coverage between genome ($Cbg) and $CDS (if     #
#         "yes", searches for CDS markers only). For more options, see README.txt.        #
#                                                                                         #
#                                                                                         #
# Output: the major output files include a text file listing queries that have exactly    #
#         one hit in each of the reference genome when the query CDS was blasted against  #
#         them, fasta files of aligned CDS sequences (optional), information of selected  #
#         intron markers (optional), and fasta files of aligned EPIC sequences (optional) #
#         for designing primers to amplify the introns.                                   #
#                                                                                         #
#                                                                                         #
# Author: Chenhong Li                                                                     #
#         School of Biological Sciences                                                   #
#         University of Nebraska - Lincoln                                                #
#         NE 68588-0118 USA                                                               #
#                                                                                         #
# Created by:  Sept. 27, 2008                                                             #
#                                                                                         #
# Last modified by:                                                                       #
#                                                                                         #
###########################################################################################

use strict;
use warnings;


use Getopt::Long;   # include the module for input
use makequery;      # module for selecting large CDS (transcripts)
use blastandparse;  # module for parsing the blast results to select for single-copy, conserved CDS
use EPIC;           # module for finding EPIC markers


my $query = "Danio_rerio"; #variable for store query species name, subsitute Danio_rerio with your taxon name
my $reference = "Gasterosteus_aculeatus Oryzias_latipes Takifugu_rubripes Tetraodon_nigroviridis"; #variable for store the reference species name, change to your taxon name
my $Lc = 100; # parameter for minimum CDS size, the default is 100
my $Li = 1000; # parameter for maximum intron size, the default is 1000
my $Iwg = 40; # within-genome blast parameter for identity, the default is 40
my $Cwg = 20; # within-genome blast paramter for coverage, the default is 20
my $Ibg = 55; # between-genome blast parameter for identity, the default is 55
my $Ibgsec = 40; #the identity for second hit in between-genome blast, used to rule out the multiple-copy genes in the reference taxa, the default is 40
my $Cbg = 70; # between-genome blast paramter for coverage, the default is 70
my $Cbgsec = 20; #coverage for the second hit, the default is 20, used to rule out the multiple-copy genes in the reference taxa, the default is 20
my $E = 0.000001; # E-value for blast, the default is 0.000001
my $pid = 85; #identity for printing EPIC. If average identity of flanking cds > $pid, print the squence alignment
my $CDS = "N"; # whether search for CDS markers only, the default is "N"ot
my $help;

my $opt = GetOptions( 'query:s', \$query,
                      'reference:s', \$reference,
                      'Lc:f', \$Lc,
                      'Li:f', \$Li,
                      'Iwg:f', \$Iwg,
                      'Cwg:f', \$Cwg,
                      'Ibg:f', \$Ibg,
                      'Cbg:f', \$Cbg,
                      'E:f', \$E,
                      'pid:f', \$pid,
                      'CDS:s', \$CDS,
                      'help!', \$help); #set command line options
                      

if (!($opt && $query && $reference) || $help) {#check for the required inputs
   print STDERR "\nExample usage:\n";
   print STDERR "\n$0 -query=\"Danio_rerio\" -reference=\"Gasterosteus_aculeatus Oryzias_latipes Takifugu_rubripes Tetraodon_nigroviridis\" [-Le] [-Li] ...\n\n";
   print STDERR "Options:\n";
   print STDERR "        -query = name of the query species you are going to\n";
   print STDERR "                 search for introns, use scientific name;\n\n";
   print STDERR "        -reference = name of the reference species used to compare with\n";
   print STDERR "                     the query species, use scientific species name;\n\n";
   print STDERR "        -Lc  = minimum length of flanking CDS; the default is 100 \n\n";
   print STDERR "        -Li  = maximum length of intron; the default is 1000 \n\n";
   print STDERR "For more optional parameters, see README.txt\n\n";
   exit;
}


#mkdir for output
my $outdir = $query . "\.results"; #name a directory for outputs
`mkdir ../$outdir -p`; #make directory for output files


#screen for large CDS
print "\nExtracting large CDS from embl files, please wait...\n";
my $nooflargecds = &makequery::getcds($query, $Lc);
print "$nooflargecds large CDS (>= $Lc bp) have been found!\n\n";


#select for single-copy CDS by within genome blast
print "Performing within-genome blast...\n";
my $numberofsingles = &blastandparse::within($query, $Iwg, $Cwg, $E); #blast for single-copy CDS
print "The within-genome blast has been done.\n";
print "There are $numberofsingles single-copy CDS.\n\n";


#extract sequences for single-copy CDS
print "Parsing single-copy CDS...\n";
&blastandparse::singleseq($query); #get the fasta sequences of single-copy CDS
print "The fasta sequences of signle-copy CDS have been extracted.\n\n";


#search for single-copy conserved CDS
print "Performing between-genome blast, please wait.\n";
my $numberofCDS = &blastandparse::between($CDS, $query, $reference, $E, $Ibg, $Ibgsec, $Cbg, $Cbgsec); #blast against refer genomes


#search for EPIC markers
if ($CDS eq "N") {#if looking for EPIC markers, or not only search for CDS
    my $numberofEPIC = &EPIC::selectEPIC($query, $reference, $Li);
    print "$numberofEPIC EPIC markers have been identified,\n";
    print "The information about EPIC markers is listed in $query.EPIClist.txt\n\n";
    
    print "Printing fasta files for selected EPIC, please wait...\n\n";
    my $epiccount = &EPIC::EPICseqfile($query, $pid);
    my $folder = $query . "_EPICmarkers";
    print "The fasta files of $epiccount EPIC (Average identity > $pid) have been saved in folder $folder.\n\n";
}

else {

    print "The selected CDS markers has been saved in $query.allCDSmarkers.txt file and $query.onehitCDSmarkers.txt file.\n";
    print "The fasta files for $numberofCDS selected CDS markers have been created!!!\n\n";
    
}
