package makequery;

###################################################################################################
#
# Moduel name: makequery
#
# This module is used to screen nuclear gene cds large than a value ($Lc) from genomic 
# sequences (EMBL flat file). Three sub routines are incluced: getcds, stack and extractseq.
#
# Datafile: a folder include all embl files of the query taxon, fasta files and index file of the
#           query genome.
#
# Imput: the query name and $Lc from main program.
#
# Output: a fasta query file including all merged CDS sequences larger than $Lc, listed in the
#         order of global position on chromosomes.
#
# Written by 
#                 Chenhong Li
#                 Univeristy of Nebraska - Lincoln, USA.
#                 Created on Sept. 2008
#                 Last modified at
#
###################################################################################################

use strict;
use warnings;


###################################################################################################
#
# sub routine: getcds
#
# This sub is used to get the CDS squences. Overlapped CDS from the same gene are merged.
#
# Imput data include the name of taxon, the size ($Lc) of selected CDS.
#
# Output file is a fasta file including list of sequences of selected CDS, which 
# will be used as the query for the within-genome blast
#
###################################################################################################

sub getcds {

    my $query = shift(@_); #get the taxa name from the main program
    my $Lc = shift(@_); #get the size criterion for CDS
    
    my ($IN_FILE, $OUT_FILE);
    my ($infilename, $outfilename);

    my %allcds;
    my $count;
    
    $outfilename = "../$query.results/$query.large_cds.fas"; #make the name of outfile
    open ($OUT_FILE, ">$outfilename") or die "Can't open the output file!!!";
    
    my $dir = "../$query.db/embl";
    opendir (DIR, $dir) or die $!; #open folder containing all embl data file
 
    while (my $file = readdir(DIR)) { #read all embl file under embl folder
        
        next if ($file =~ /^\./); #skip files beginning with .
        print "$file\n"; #print the file under processing to the screen

        $infilename = "../$query.db/embl/$file"; #get the name of embl file
        open ($IN_FILE, "<$infilename") or die "Can't open the embl file!!!";


        my $ac;
        my @cdspos;
        my ($isgeneflag, $iscdsflag, $isnoteflag, $secondgeneflag);
        my ($genename, $locustag, $note, $strand);
    
        while (my $line = readline($IN_FILE)) { #read the embl file
            chomp $line;

            if ($line =~ /^AC\s+(.+)/) { #extract the ac number
                $ac = $1;
                $secondgeneflag = 0; #set the flag to skip first "gene"
                next;
            }
    
            if ($isgeneflag) { #if it is reading the gene information
                if ($line =~ /^FT\s+\/gene=(.+)/) { #extract gene name
                    $genename = $1;
                }
                elsif ($line =~ /^FT\s+\/locus_tag=(.+)/) { #extract locus tag
                    $locustag = $1; 
                }           
                elsif ($line =~ /^FT\s+\/note=(.+)/) { #get the note and set isnoteflag
                    $note = $1;
                    if ($line !~ /"$/) { #if not the end of notes
                        $isnoteflag = 1;
                    }
                }
                elsif ($isnoteflag) { #if continue reading note
                    my ($morenote) = $line =~ /^FT\s+(.+)/;
                    $note = $note . " " . $morenote;
                    if ($line =~ /"$/) { #if the end of note
                        $isnoteflag = 0; #set noteflag back to zero
                    }
                }
                elsif (($line =~ /^FT\s+mRNA/) || ($line =~ /^FT\s+misc_RNA/)) {
                    $isgeneflag = 0; #close reading gene
                }
            }
    
            if ($line =~ /^FT\s+CDS\s+/) { #check if start CDS
                my @tmp1 = $line =~ /(\d+\.\.\d+)/g; #extract position of CDS
                push @cdspos, @tmp1;
                if ($line =~ /,$/) { #if end with ",", there is more CDS, set $iscdsflag to 1
                    $iscdsflag = 1;
                }
                if ($line =~ /\d+$/) { #if end with digit, set $iscdsflag to 0
                    $iscdsflag = 0;
                }
                next;
            }
    
            if ($iscdsflag) {
                my @tmp2 = $line =~ /(\d+\.\.\d+)/g;
                push @cdspos, @tmp2;
                if ($line =~ /\)$/) {#if end with parenthesis, set $iscdsflag to 0
                    $iscdsflag = 0;
                }
            }

            if ($line =~ /^FT\s{3}gene\s+/) { #start a new gene

                if ($secondgeneflag) { #if after start the first gene

                    if (scalar(@cdspos)) { #if there are CDS

                        foreach my $cds (@{stack(\@cdspos, $Lc)}) { #call sub stack to merge CDS and screen by size $Lc

                            my ($startpos, $endpos) = split/\.\./, $cds; #get the start and ending pos of CDS
                            my @acnumber = split/:/, $ac; #get the global pos
                            my $globalstart = $startpos + $acnumber[3] - 1;
                            my $globalend = $endpos + $acnumber[3] -1;
                            my ($newac, $pos1, $pos2);
                            if ($strand == 1) {#if + strand, put smaller pos infront of larger one
                                $newac = join(':', $acnumber[0], $acnumber[1], $acnumber[2], $globalstart, $globalend);
                                $pos1 = $globalstart - 1;# get the position for extracting sequence later
                                $pos2 = $globalend - 1;
                            }
                            if ($strand == -1) {# if - strand, put larger pos in the front
                                $newac = join(':', $acnumber[0], $acnumber[1], $acnumber[2], $globalend, $globalstart);
                                $pos1 = $globalend - 1;
                                $pos2 = $globalstart - 1;
                            }
                            my $newid = "$newac|$genename|$locustag|$strand|$note";
                            $newid =~ s/ /=/g; #substitue the space with "=", cause fasta id doesn't allow space
                            $newid =~ s/\"//g; #delete " in the gene note
                            
                            if (($newac !~ /^\w+:\w+:c/) && ($newac !~ /^\w+:\w+:MT:.+/)) { #if the AC doesn't match duplicated entry of homo_sapeins or mitochondrail sequences
                                #get the sequence for the cds
                                my $seqfile = "../$query.db/$query.genome.new.fas";
                                my $indexfile = "../$query.db/$query.genome.new.index";
                                my ($seq, $strain) = &makequery::extractseq($acnumber[2], $pos1, $pos2, $seqfile, $indexfile); #use sub extractseq to get the sequences

                                $allcds{$newid} = $seq;
                            }
                        
                        }                    
                    
                    }
                
                }
                
                $secondgeneflag = 1;
                @cdspos = ();
                $isgeneflag = 1;
                $genename = "";
                $locustag = "";
                $note = "";
                if ($line =~ /complement/g) { #check if the next gene is on the complement strand
                    $strand = -1;
                }
                else {
                    $strand = 1;
                }
            }
            

            if ($line =~ /^\/\//) { #if the end of the entry
            
                #first we need add the information for the last CDS of the gene
                if (scalar(@cdspos)) {   
                    
                    foreach my $cds (@{stack(\@cdspos, $Lc)}) { #call sub stack to merge CDS and screen by size $Lc
                        
                        my ($startpos, $endpos) = split/\.\./, $cds; #get the start and ending pos of CDS
                        my @acnumber = split/:/, $ac; #get the global pos
                        my $globalstart = $startpos + $acnumber[3] - 1;
                        my $globalend = $endpos + $acnumber[3] - 1;
                        my ($newac, $pos1, $pos2);
                        if ($strand == 1) {#if + strand, smaller pos infront of larger one
                            $newac = join(':', $acnumber[0], $acnumber[1], $acnumber[2], $globalstart, $globalend);
                            $pos1 = $globalstart - 1;
                            $pos2 = $globalend - 1;
                        }
                        if ($strand == -1) {# if - strand, larger num in the front
                            $newac = join(':', $acnumber[0], $acnumber[1], $acnumber[2], $globalend, $globalstart);
                            $pos1 = $globalend - 1;
                            $pos2 = $globalstart - 1;
                        }
                        my $newid = "$newac|$genename|$locustag|$strand|$note";
                        $newid =~ s/\s/=/g; #substitue the space with "=", cause fasta id doesn't allow space
                        $newid =~ s/\"//g; #delete " in the gene note
                        
                        if (($newac !~ /^\w+:\w+:c/) && ($newac !~ /^\w+:\w+:MT:.+/)) { #if the AC doesn't match duplicated entry of homo_sapeins or mitochondrail sequences
                            
                            #get the sequence for the cds
                            my $seqfile = "../$query.db/$query.genome.new.fas";
                            my $indexfile = "../$query.db/$query.genome.new.index";
                            my ($seq, $strain) = &makequery::extractseq($acnumber[2], $pos1, $pos2, $seqfile, $indexfile); #use sub extractseq to get the sequences
                        
                            $allcds{$newid} = $seq;
                        }
                        
                    }
                
                }
                
            
                #reset for the next entry
                @cdspos = ();
                $genename = "";
                $locustag = "";
                $note = "";
            
            }#return to next entry
        
         
        }#returen to next dir file
        
    
        close ($IN_FILE) or die "Can't close the embl file!!!";
    
        
    }#end of all dir
    
    
    
    #now we can print out all information saved in %allcds for this entry (under one ac number)
    my @cdskey = keys(%allcds);
                
    #sort by the chromosome id then by the global position of the cds
    my @sortedkey = sort mysort @cdskey;
    
    sub mysort { 
        my $aa = $a;
        my $bb = $b;
        ($aa =~ /^\w+:\w*:(\w+):\d+:\d+\|\w+\|/)[0] cmp ($bb =~ /^\w+:\w*:(\w+):\d+:\d+\|\w+\|/)[0] || ($a =~ /^\w+:\w*:\w+:(\d+):\d+\|\w+\|/)[0] <=> ($b =~ /^\w+:\w*:\w+:(\d+):\d+\|\w+\|/)[0];
    }             
                
                
    foreach my $cdsid (@sortedkey) {
                       
        print $OUT_FILE ">$cdsid\n"; #print the sequence id
        print $OUT_FILE "$allcds{$cdsid}\n"; #print the sequence
        $count ++;
                       
    }
        

    closedir (DIR);
    close ($OUT_FILE) or die "Can't close the CDS output file!!!";

    return($count);

}



###################################################################################################
#
# sub routine: stack
#
# This sub is used to merge the coding sequences (CDS) of the same gene
#
# Imput: the array of start and end positions of CDS from all transcripts of the same gene, size 
# for selecting exon ($Lc).
#
# Output: an array of start and ending positions of merged CDS whose size are less than $Lc
#
###################################################################################################

sub stack {

    my $cdsref = shift(@_); #get the ref for the array of CDS of all transcripts
    my @originalcds = @$cdsref; #deref
    my $Lc = shift (@_); #get the size criteron for selecting exons
    
    my @sortedcds = sort {
        ($a =~ /(\d+)\./)[0] <=> ($b =~ /(\d+)\./)[0]
    } @originalcds; #sort the CDS position by the start position

    my @mergedcds = ();
    my ($startpos, $endpos, $startlag, $endlag);
    foreach my $cds (@sortedcds) {
        ($startpos, $endpos) = split /\.\./, $cds; #get the start and ending pos for each CDS
        if (!($endlag)) { #if it is the first CDS of the gene, record the pos and go next
            $startlag = $startpos;
            $endlag = $endpos;
            next;
        } 
        elsif ($startpos <= $endlag) { #if the start pos is smaller than the ending pos of the lag CDS
            if ($endpos >= $endlag) { #use the large ending pos and merge the two CDS
                $endlag = $endpos;
            }
            next;
        } 
        else { #if the start pos is larger than the ending pos of the lag CDS, push the lag CDS
            if (($endlag - $startlag) >= $Lc) { #if the CDS is larger than $Lc
                push @mergedcds, "$startlag..$endlag";
            }
            $startlag = $startpos;
            $endlag = $endpos;
        }
    }

    #need to push the last CDS
    if (($endlag - $startlag) >= $Lc) { #if the CDS is larger than $Lc
        push @mergedcds, "$startlag..$endlag";
    }

    return (\@mergedcds); #return ref of the new array of merged CDS

}    



###################################################################################################
#
# sub routine: extractseq
#
# This sub is used to extract the sequence from large fasta sequences.
#
# Imput data include the $seqid, $pos1 and $pos2.
#
# Return is the extracted sequence and the strain (+ or -).
#
###################################################################################################

sub extractseq {

    my $seqid = shift(@_); #get the $seqid from main program
    my $pos1 = shift(@_); #get the position 1
    my $pos2 = shift(@_); #get the position 2
    my $seqfile = shift(@_); #get the large fasta file name
    my $indexfile = shift(@_); #get the index file name

    my $seqlength = abs($pos1 - $pos2) + 1; #calculate the length of the seq
    
    my $strain = 1; 
    my $seqstart;
    
    if ($pos1 < $pos2) { #pick the small number as sequence start point
       $seqstart = $pos1;
    }
    elsif ($pos1 > $pos2) {
        $seqstart = $pos2;
        $strain = -1;
    }
    

    my ($SEQ_FILE, $INDEX_FILE);
    open ($SEQ_FILE, "<$seqfile") or die "Cannot open $seqfile for reading ($!)";
    open ($INDEX_FILE, "<$indexfile") or die "Cannot open $indexfile for reading ($!)";


    # First read the index file
    my %index;
    my $id;
    while (my $line = readline ($INDEX_FILE)) {
    	chomp $line;
    	my ($wholeid, $position) = split /\t/, $line;# the id in index file is longer than in embl file
    	($id) = $wholeid =~ /^(\S+).*/;
    	$index{$id} = $position;
    }


    # and now we have superfast access to our sequences:
    seek $SEQ_FILE, $index{$seqid}, 0;  # lets jump to our seq
    readline($SEQ_FILE);  # first line is header, skip
    my $seq = readline($SEQ_FILE);	# second line is the whole sequence
    my $subseq = substr($seq, $seqstart, $seqlength); #get the substring

    if ($strain == -1) { #reverse and complement the sequence if minus strain
        my $revsubseq = reverse $subseq;
        $revsubseq =~ tr/ACTGN/TGACN/;
        $subseq = $revsubseq;
    }

    close ($SEQ_FILE) or die "Can't close the new file!!!";
    close ($INDEX_FILE) or die "Can't close the index file!!!";

    return($subseq, $strain);
}







1
