package makequerygff;

###################################################################################################
#
# Moduel name: makequerygff
#
# This module is used to screen nuclear gene cds large than a value ($Lc) from genomic 
# sequences (gff file). Three sub routines are incluced: getcds, stack and extractseq.
#
# Datafile: a folder include all gff files of the query taxon, fasta files and index file of the
#           query genome.
#
# Imput: the query name and $Lc from main program.
#
# Output: a fasta query file including all merged CDS sequences larger than $Lc, listed in the
#         order of global position on chromosomes.
#
# Written by 
#                 Chenhong Li
#                 Univeristy of Nebraska - Lincoln, USA.
#                 Created on June 2010
#                 Last modified at
#
###################################################################################################

use strict;
use warnings;


###################################################################################################
#
# sub routine: getcds
#
# This sub is used to get the CDS squences. Overlapped CDS from the same gene are merged.
#
# Imput data include the name of taxon, the size ($Lc) of selected CDS.
#
# Output file is a fasta file including list of sequences of selected CDS, which 
# will be used as the query for the within-genome blast
#
###################################################################################################

sub getcds {

    my $query = shift(@_); #get the taxa name from the main program
    my $Lc = shift(@_); #get the size criterion for CDS
    
    my ($IN_FILE, $OUT_FILE);
    my ($infilename, $outfilename);

    my %allcds;
    my $count;
    
    $outfilename = "../$query.results/$query.large_cds.fas"; #make the name of outfile
    open ($OUT_FILE, ">$outfilename") or die "Can't open the output file!!!";
    
    my $dir = "../$query.db/gff";
    opendir (DIR, $dir) or die $!; #open folder containing all gff data file
 
    while (my $file = readdir(DIR)) { #read all gff file under gff folder
        
        next if ($file =~ /^\./); #skip files beginning with .
        print "$file\n"; #print the file under processing to the screen

        $infilename = "../$query.db/gff/$file"; #get the name of gff file
        open ($IN_FILE, "<$infilename") or die "Can't open the gff file!!!";


        my @cdspos;
        my ($scaffold, $pos1, $pos2, $source, $genename);
        my $strand = "strand";
        my $genenamelag = "genenamelag";
        my $strandlag = "strandlag";
        my $scaffoldlag = "scaffoldlag";
        my $sourcelag;
    
        while (my $line = readline($IN_FILE)) { #read the gff file
            chomp $line;

            if (($scaffold, $source, $pos1, $pos2, $strand, $genename) = $line =~ /^(\S+)\s+(\S+)\s+CDS\s+(\d+)\s+(\d+)\s+\S+\s+([\+\-]{1})\s+\S+\s+(.+)/) { #extract the scaffold number CDS position and the strand information

                if (((($strandlag ne "strandlag") &&  ($strandlag ne $strand)) || (($scaffoldlag ne "scaffoldlag") &&  ($scaffoldlag ne $scaffold))) || (eof)) { #if start on a different strand, probably a new gene, or a different scaffold

                    if (scalar(@cdspos)) {#if there is values in the array
                    
                        foreach my $cds (@{stack(\@cdspos, $Lc)}) { #call sub stack to merge CDS and screen by size $Lc
                        
                            my ($startpos, $endpos) = split/\.\./, $cds; #get the start and ending pos of CDS
                            my ($newac, $p1, $p2);
                            if ($strandlag eq "+") {#if + strand, smaller pos infront of larger one
                                $newac = join(':', $sourcelag, $query, $scaffoldlag, $startpos, $endpos);
                                $p1 = $startpos - 1;
                                $p2 = $endpos - 1;
                            }
                            if ($strandlag eq "-") {# if - strand, larger num in the front
                                $newac = join(':', $sourcelag, $query, $scaffoldlag, $endpos, $startpos);
                                $p1 = $endpos - 1;
                                $p2 = $startpos - 1;
                            }
                            my $newid = "$newac|$genenamelag|locustag|$strandlag|note";
                            $newid =~ s/\s/=/g; #substitue the space with "=", cause fasta id doesn't allow space
                            $newid =~ s/\"//g; #delete " in the gene note
                        
                            if (($newac !~ /^\w+:\w+:c/) && ($newac !~ /^\w+:\w+:MT:.+/)) { #if the AC doesn't match duplicated entry of homo_sapeins or mitochondrail sequences
                                
                                #get the sequence for the cds
                                my $seqfile = "../$query.db/$query.genome.new.fas";
                                my $indexfile = "../$query.db/$query.genome.new.index";
                                my ($seq, $strandread) = &makequerygff::extractseq($scaffoldlag, $p1, $p2, $seqfile, $indexfile); #use sub extractseq to get the sequences
                        
                                $allcds{$newid} = $seq;
                            }
                                my $temp1 = $pos1 . ".." . $pos2;
                                push (@cdspos, $temp1);
                        }
                    }

                
                    #reset for the next entry
                    @cdspos = ();
                
                }
                
                else {

                    my $temp1 = $pos1 . ".." . $pos2;
                    push (@cdspos, $temp1);
                
                }
                
                $genenamelag = $genename;
                $scaffoldlag = $scaffold;
                $sourcelag = $source;
                $strandlag = $strand; #tracking the strand
                
            }
            
            elsif (eof) {#if the end of file
                
                    
                if (scalar(@cdspos)) {#if there is values in the array
                    
                    foreach my $cds (@{stack(\@cdspos, $Lc)}) { #call sub stack to merge CDS and screen by size $Lc
                    
                        my ($startpos, $endpos) = split/\.\./, $cds; #get the start and ending pos of CDS
                        my ($newac, $p1, $p2);
                        if ($strandlag eq "+") {#if + strand, smaller pos infront of larger one
                            $newac = join(':', $sourcelag, $query, $scaffoldlag, $startpos, $endpos);
                            $p1 = $startpos - 1;
                            $p2 = $endpos - 1;
                        }
                        if ($strandlag eq "-") {# if - strand, larger num in the front
                            $newac = join(':', $sourcelag, $query, $scaffoldlag, $endpos, $startpos);
                            $p1 = $startpos - 1;
                            $p2 = $endpos - 1;
                        }
                        my $newid = "$newac|$genenamelag|locustag|$strandlag|note";
                        $newid =~ s/\s/=/g; #substitue the space with "=", cause fasta id doesn't allow space
                        $newid =~ s/\"//g; #delete " in the gene note
                        
                        if (($newac !~ /^\w+:\w+:c/) && ($newac !~ /^\w+:\w+:MT:.+/)) { #if the AC doesn't match duplicated entry of homo_sapeins or mitochondrail sequences
                                
                            #get the sequence for the cds
                            my $seqfile = "../$query.db/$query.genome.new.fas";
                            my $indexfile = "../$query.db/$query.genome.new.index";
                            my ($seq, $strandread) = &makequerygff::extractseq($scaffoldlag, $p1, $p2, $seqfile, $indexfile); #use sub extractseq to get the sequences
                        
                            $allcds{$newid} = $seq;
                        }
                    }
                }
            }
            
        }#return to next line
        
        close ($IN_FILE) or die "Can't close the gff file!!!";
         
    }#return to next dir file
        
    
    
        
    
    
    
    #now we can print out all information saved in %allcds
    my @cdskey = keys(%allcds);
                
    #sort by the scaffold id then by the global position of the cds
    my @sortedkey = sort mysort @cdskey;
    
    sub mysort { 
        my $aa = $a;
        my $bb = $b;
        ($aa =~ /^\S*:\S*:(\S+):\d+:\d+\|\S*\|\S*\|\S*\|\S*/)[0] cmp ($bb =~ /^\S*:\S*:(\S+):\d+:\d+\|\S*\|\S*\|\S*\|\S*/)[0] || ($a =~ /^\S*:\S*:\S+:(\d+):\d+\|\S*\|\S*\|\S*\|\S*/)[0] <=> ($b =~ /^\S*:\S*:\S+:(\d+):\d+\|\S*\|\S*\|\S*\|\S*/)[0];
    }             
                
                
    foreach my $cdsid (@sortedkey) {
                       
        print $OUT_FILE ">$cdsid\n"; #print the sequence id
        print $OUT_FILE "$allcds{$cdsid}\n"; #print the sequence
        $count ++;
                       
    }
        

    closedir (DIR);
    close ($OUT_FILE) or die "Can't close the CDS output file!!!";

    return($count);

}


###################################################################################################
#
# sub routine: stack
#
# This sub is used to merge the coding sequences (CDS) of the same gene
#
# Imput: the array of start and end positions of CDS from all transcripts of the same gene, size 
# for selecting exon ($Lc).
#
# Output: an array of start and ending positions of merged CDS whose size are less than $Lc
#
###################################################################################################

sub stack {

    my $cdsref = shift(@_); #get the ref for the array of CDS of all transcripts
    my @originalcds = @$cdsref; #deref
    my $Lc = shift (@_); #get the size criteron for selecting exons
    
    my @sortedcds = sort {
        ($a =~ /(\d+)\./)[0] <=> ($b =~ /(\d+)\./)[0]
    } @originalcds; #sort the CDS position by the start position

    my @mergedcds = ();
    my ($startpos, $endpos, $startlag, $endlag);
    foreach my $cds (@sortedcds) {
        ($startpos, $endpos) = split /\.\./, $cds; #get the start and ending pos for each CDS
        if (!($endlag)) { #if it is the first CDS of the gene, record the pos and go next
            $startlag = $startpos;
            $endlag = $endpos;
            next;
        } 
        elsif ($startpos <= $endlag) { #if the start pos is smaller than the ending pos of the lag CDS
            if ($endpos >= $endlag) { #use the large ending pos and merge the two CDS
                $endlag = $endpos;
            }
            next;
        } 
        else { #if the start pos is larger than the ending pos of the lag CDS, push the lag CDS
            if (($endlag - $startlag) >= $Lc) { #if the CDS is larger than $Lc
                push @mergedcds, "$startlag..$endlag";
            }
            $startlag = $startpos;
            $endlag = $endpos;
        }
    }

    #need to push the last CDS
    if (($endlag - $startlag) >= $Lc) { #if the CDS is larger than $Lc
        push @mergedcds, "$startlag..$endlag";
    }

    return (\@mergedcds); #return ref of the new array of merged CDS

}    



###################################################################################################
#
# sub routine: extractseq
#
# This sub is used to extract the sequence from large fasta sequences.
#
# Imput data include the $seqid, $pos1 and $pos2.
#
# Return is the extracted sequence and the strand (+ or -).
#
###################################################################################################

sub extractseq {

    my $seqid = shift(@_); #get the $seqid from main program
    my $pos1 = shift(@_); #get the position 1
    my $pos2 = shift(@_); #get the position 2
    my $seqfile = shift(@_); #get the large fasta file name
    my $indexfile = shift(@_); #get the index file name

    my $seqlength = abs($pos1 - $pos2) + 1; #calculate the length of the seq
    
    my $strand = 1; 
    my $seqstart;
    
    if ($pos1 < $pos2) { #pick the small number as sequence start point
       $seqstart = $pos1;
    }
    elsif ($pos1 > $pos2) {
        $seqstart = $pos2;
        $strand = -1;
    }
    

    my ($SEQ_FILE, $INDEX_FILE);
    open ($SEQ_FILE, "<$seqfile") or die "Cannot open $seqfile for reading ($!)";
    open ($INDEX_FILE, "<$indexfile") or die "Cannot open $indexfile for reading ($!)";


    # First read the index file
    my %index;
    my $id;
    while (my $line = readline ($INDEX_FILE)) {
    	chomp $line;
    	my ($wholeid, $position) = split /\t/, $line;# the id in index file is longer than in gff file
    	($id) = $wholeid =~ /^(\S+).*/;
    	$index{$id} = $position;
    }


    # and now we have superfast access to our sequences:
    seek $SEQ_FILE, $index{$seqid}, 0;  # lets jump to our seq
    readline($SEQ_FILE);  # first line is header, skip
    my $seq = readline($SEQ_FILE);	# second line is the whole sequence
    my $subseq = substr($seq, $seqstart, $seqlength); #get the substring

    if ($strand == -1) { #reverse and complement the sequence if minus strand
        my $revsubseq = reverse $subseq;
        $revsubseq =~ tr/ACTGN/TGACN/;
        $subseq = $revsubseq;
    }

    close ($SEQ_FILE) or die "Can't close the new file!!!";
    close ($INDEX_FILE) or die "Can't close the index file!!!";

    return($subseq, $strand);
}


1
