#!/arch/bin/perl -w

###################################################################################################
#
# Name: singlelarge.pl
#
# This script is used to screen nuclear gene cds large than a value ($Lc) from genome (by parsing 
# embl flat file or gff file), blast the genome sequence of itself to find the single copy gene
# and extract the fasta sequences.
#
# Datafile: fasta files and index file of the query genome, local blast data base, a folder include
# all embl or gff files of the query taxon.
#
# Imput: the query name and $Lc (minimum CDS size), $Iwg (withing-genome blast identity), $Cwg
# (within genome blast coverage), $E (E-value for blast) from main program, $gff (switch of
# gff/EMBL files).
#
# Output: a fasta file including all merged CDS sequences larger than $Lc and are single copy,
#         listed in the order of global position on chromosomes or scaffolds.
#
# Written by 
#                 Chenhong Li
#                 Univeristy of Nebraska - Lincoln, USA.
#                 Created on June 2010
#                 Last modified on
#
###################################################################################################

use strict;
use warnings;


use Getopt::Long;   # include the module for input
use makequery;      # module for selecting large CDS (coding sequences) from embl file
use makequerygff;   # module for selecting large CDS (coding sequences) from gff file
use blastandparse;  # module for parsing the blast results to select for single-copy, conserved CDS


my $query; #variable for store query species name, subsitute Danio_rerio with your taxon name
my $gff = "N"; #whether the data format is gff or not, the default is "N"ot
my $Lc = 100; # parameter for minimum CDS size, the default is 100
my $Iwg = 40; # within-genome blast parameter for identity, the default is 40
my $Cwg = 20; # within-genome blast paramter for coverage, the default is 20
my $E = 0.000001; # E-value for blast, the default is 0.000001
my $help;

my $opt = GetOptions( 'query:s', \$query,
                      'gff:s', \$gff,
                      'Lc:f', \$Lc,
                      'Iwg:f', \$Iwg,
                      'Cwg:f', \$Cwg,
                      'E:f', \$E,
                      'help!', \$help); #set command line options
                      

if (!($opt && $query) || $help) {#check for the required inputs
   print STDERR "\nExample usage:\n";
   print STDERR "\n$0 -query=\"Danio_rerio\" [-Le] ...\n\n";
   print STDERR "Options:\n";
   print STDERR "        -query = name of the query species you are going to\n";
   print STDERR "                 search large and single-copy CDS, use scientific name;\n\n";
   print STDERR "        -gff = if it is gff file, the default is no\n";
   print STDERR "        -Lc  = minimum length of CDS; the default is 100 \n\n";
   print STDERR "For more optional parameters, see README.txt\n\n";
   exit;
}


#mkdir for output
my $outdir = $query . "\.results"; #name a directory for outputs
`mkdir ../$outdir -p`; #make directory for output files


#screen for large CDS
print "\nExtracting large CDS from embl/gff files, please wait...\n";
my $nooflargecds;
if ($gff eq "N") {
    $nooflargecds = &makequery::getcds($query, $Lc);
} else {
    $nooflargecds = &makequerygff::getcds($query, $Lc);
}
print "$nooflargecds large CDS (>= $Lc bp) have been found!\n\n";


#select for single-copy CDS by within genome blast
print "Performing within-genome blast...\n";
my $numberofsingles = &blastandparse::within($query, $Iwg, $Cwg, $E); #blast for single-copy CDS
print "The within-genome blast has been done.\n";
print "There are $numberofsingles single-copy CDS.\n\n";


#extract sequences for single-copy CDS
print "Parsing single-copy CDS...\n";
&blastandparse::singleseq($query); #get the fasta sequences of single-copy CDS
print "The fasta sequences of signle-copy CDS have been extracted.\n\n";
